/**
 * PETALS: PETALS Services Platform Copyright (C) 2005 EBM WebSourcing
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 * 
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: AutoLoaderServiceImpl.java,v 1.2 2006/04/07 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------
 */

package org.ow2.petals.jbi.management.autoload;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Timer;

import javax.management.InstanceNotFoundException;
import javax.management.MBeanException;
import javax.management.MBeanServer;
import javax.management.ObjectName;
import javax.management.ReflectionException;

import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.communication.jmx.JMXService;
import org.ow2.petals.container.lifecycle.ServiceAssemblyLifeCycle;
import org.ow2.petals.container.lifecycle.ServiceUnitLifeCycle;
import org.ow2.petals.jbi.descriptor.JBIDescriptorException;
import org.ow2.petals.jbi.descriptor.original.generated.Jbi;
import org.ow2.petals.jbi.management.admin.AdminService;
import org.ow2.petals.jbi.management.deployment.DeploymentServiceMBean;
import org.ow2.petals.jbi.management.installation.InstallationServiceMBean;
import org.ow2.petals.jbi.management.util.PackageHelper;
import org.ow2.petals.jbi.management.util.XMLResult;
import org.ow2.petals.kernel.api.server.PetalsException;
import org.ow2.petals.kernel.configuration.ConfigurationService;
import org.ow2.petals.kernel.configuration.ContainerConfiguration;
import org.ow2.petals.system.repository.RepositoryService;
import org.ow2.petals.system.repository.RepositoryServiceImpl.EntityType;
import org.ow2.petals.system.systemstate.SystemStateService;
import org.ow2.petals.systemstate.generated.Component;
import org.ow2.petals.systemstate.generated.ServiceAssembly;
import org.ow2.petals.systemstate.generated.SharedLibrary;
import org.ow2.petals.util.LoggingUtil;

import static org.ow2.petals.kernel.api.server.PetalsServer.INSTALLED_DIRECTORY;
import static org.ow2.petals.kernel.api.server.PetalsServer.INSTALL_DIRECTORY;

/**
 * This class is used to load automatically components which are placed in the
 * "autoLoad" directory. It scans the directory periodically and check if new
 * components are presents. The @ link InstallationServiceMBean} is called for
 * each of those components. The whole process of installation is done (install
 * + start) After being threat, the archives of the components are removed. A
 * file "fileName.success" will be created if success A file "fileName.failed"
 * will be created otherwise, containing the related exception
 * 
 * @author alouis
 * 
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.ow2.petals.jbi.management.autoload.AutoLoaderService.class))
public class AutoLoaderServiceImpl implements AutoLoaderService {

    private static File installDirectory;

    private static File installedDirectory;

    private static final long scanPeriod = 4000L;

    @Requires(name = "deployment", signature = org.ow2.petals.jbi.management.deployment.DeploymentServiceMBean.class)
    protected DeploymentServiceMBean deploymentService;

    @Requires(name = "installation", signature = org.ow2.petals.jbi.management.installation.InstallationServiceMBean.class)
    protected InstallationServiceMBean installationService;

    @Requires(name = "jmx", signature = org.ow2.petals.communication.jmx.JMXService.class)
    protected JMXService jmxService;

    protected LoggingUtil log;

    @Monolog(name = "logger")
    protected Logger logger;

    @Requires(name = "repository", signature = org.ow2.petals.system.repository.RepositoryService.class)
    protected RepositoryService repositoryService;

    @Requires(name = "systemstate", signature = org.ow2.petals.system.systemstate.SystemStateService.class)
    protected SystemStateService systemStateService;

    @Requires(name = "admin", signature = org.ow2.petals.jbi.management.admin.AdminService.class)
    private AdminService adminService;

    @Requires(name = "configuration", signature = org.ow2.petals.kernel.configuration.ConfigurationService.class)
    private ConfigurationService configurationService;

    /**
     * The local container information.
     */
    private ContainerConfiguration containerConfiguration;

    private Timer installDirectoryScanTimer;

    private Timer installedDirectoryScanTimer;

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.jbi.management.autoload.AutoLoaderService#install(java
     * .util.List)
     */
    public void install(List<File> archiveFiles) {
        this.log.start();

        final List<File> slToInstall = this.getJBIArchives(EntityType.SL_TYPE, archiveFiles);
        for (final File archive : slToInstall) {
            try {
                this.performInstallSL(archive);
            } catch (final Throwable e) {
                this.log.error("Failed to auto-install of a shared library", e);
            } finally {
                try {
                    archive.delete();
                } catch (final SecurityException e) {
                    this.log.error(e.getMessage(), e);
                }
            }
        }
        archiveFiles.removeAll(slToInstall);

        final List<File> compToInstall = this.getJBIArchives(EntityType.COMPONENT_TYPE,
                archiveFiles);
        for (final File archive : compToInstall) {
            try {
                this.performInstallComponent(archive);
            } catch (final Throwable e) {
                this.log.error("Error during the auto- installation of a component", e);
            } finally {
                try {
                    archive.delete();
                } catch (final SecurityException e) {
                    this.log.error(e.getMessage(), e);
                }
            }
        }
        archiveFiles.removeAll(compToInstall);

        final List<File> saToInstall = this.getJBIArchives(EntityType.SA_TYPE, archiveFiles);
        for (final File archive : saToInstall) {
            try {
                this.performDeploySA(archive);
            } catch (final Throwable e) {
                this.log.error("Error during the deployment of a service assembly: ", e);
            } finally {
                try {
                    archive.delete();
                } catch (final SecurityException e) {
                    this.log.error(e.getMessage(), e);
                }
            }
        }
        archiveFiles.removeAll(saToInstall);

        // clean the remaining archives
        for (final File archive : archiveFiles) {
            try {
                this.log.warning("The archive '" + archive.getName()
                        + "' remains in the install list. PEtALS deletes it");
                archive.delete();
            } catch (final SecurityException e) {
                this.log.error(e.getMessage(), e);
            }
        }

        this.log.end();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.jbi.management.autoload.AutoLoaderService#uninstall(java
     * .util.List)
     */
    public void uninstall(List<File> archiveFiles) {
        this.log.call();

        try {
            // Check if one of the files to uninstall match one of the
            // archiveUrl of the installed entities. If yes, this entity must be
            // uninstalled.
            final List<String> serviceAssemblyToUninstall = this
                    .getServiceAssemblyNames(archiveFiles);
            final List<String> componentToUninstall = this.getComponentNames(archiveFiles);
            final List<String> sharedLibraryToUninstall = this.getSharedLibraryNames(archiveFiles);

            for (final String name : serviceAssemblyToUninstall) {
                this.performUndeploySA(name);
            }
            for (final String name : componentToUninstall) {
                this.performUninstallComponent(name);
            }
            for (final String name : sharedLibraryToUninstall) {
                this.performUninstallSL(name);
            }
        } catch (final Throwable e) {
            this.log.error("Error during uninstallation of JBI artifact(s)", e);
        }

    }

    /**
     * For each JBI artefact to uninstall, search the dependent artefacts and
     * add them to the List if they are not already set.
     * 
     * @param filesToUninstall
     *            The initial collection of files to uninstall
     */
    protected void resolveJBIArtefactDependencies(final List<File> archivesToUninstall) {

        // resolve the dependent SAs against the components to uninstall
        final List<String> componentsToUninstall = this.getComponentNames(archivesToUninstall);
        if (!componentsToUninstall.isEmpty()) {
            final Collection<ServiceAssemblyLifeCycle> serviceAssemblyLifeCycles = this.adminService
                    .getServiceAssemblies().values();
            for (final String componentName : componentsToUninstall) {
                for (final ServiceAssemblyLifeCycle serviceAssemblyLifeCycle : serviceAssemblyLifeCycles) {
                    for (final ServiceUnitLifeCycle serviceUnitLifeCycle : serviceAssemblyLifeCycle
                            .getServiceUnitLifeCycles()) {
                        if (componentName.equals(serviceUnitLifeCycle.getTargetComponentName())) {
                            final String serviceAssemblyName = serviceAssemblyLifeCycle
                                    .getServiceAssembly().getIdentification().getName();
                            final String archiveName = this.systemStateService
                                    .getServiceAssemblyStateHolder(serviceAssemblyName)
                                    .getArchiveName();
                            final File archiveFile = new File(installedDirectory, archiveName);
                            if (!archivesToUninstall.contains(archiveFile)) {
                                archivesToUninstall.add(archiveFile);
                            }
                            break;
                        }
                    }
                }
            }
        }

        // resolve the dependent components against the SLs to uninstall
        final List<String> sharedLibrariesToUninstall = this
                .getSharedLibraryNames(archivesToUninstall);
        for (final String sharedLibraryName : sharedLibrariesToUninstall) {
            try {
                String[] componentNames = this.installationService
                        .getInstalledComponentsForSharedLibrary(sharedLibraryName);
                for (String componentName : componentNames) {
                    final String archiveName = this.systemStateService
                            .getServiceAssemblyStateHolder(componentName).getArchiveName();
                    final File archiveFile = new File(installedDirectory, archiveName);
                    if (!archivesToUninstall.contains(archiveFile)) {
                        archivesToUninstall.add(archiveFile);
                    }
                }
            } catch (Exception e) {
                continue;
            }
        }
    }

    /**
     * Initialize the autoloader.
     */
    @LifeCycle(on = LifeCycleType.START)
    protected void start() throws IOException {
        this.log = new LoggingUtil(this.logger);
        this.log.call();

        this.containerConfiguration = this.configurationService.getContainerConfiguration();
        installDirectory = new File(this.containerConfiguration.getRootDirectoryPath(),
                INSTALL_DIRECTORY);
        installedDirectory = new File(this.containerConfiguration.getRootDirectoryPath(),
                INSTALLED_DIRECTORY);

        // old start method
        this.installDirectoryScanTimer = new Timer();
        this.installedDirectoryScanTimer = new Timer();
        final InstallDirectoryScanner installScan = new InstallDirectoryScanner(this,
                installDirectory);
        final InstalledDirectoryScanner installedScan = new InstalledDirectoryScanner(this,
                installedDirectory);
        this.installDirectoryScanTimer.schedule(installScan, scanPeriod, scanPeriod);
        this.installedDirectoryScanTimer.schedule(installedScan, scanPeriod, scanPeriod);
    }

    /**
     * Stop the autoload of the components. TODO : synchronization, does the
     * autoload must wait before the end of an installation before closing?
     */
    @LifeCycle(on = LifeCycleType.STOP)
    protected void stop() {
        this.log.call();

        this.installDirectoryScanTimer.cancel();
        this.installedDirectoryScanTimer.cancel();
    }

    /**
     * Get the component names from the archive list.
     * 
     * @param archivesToUninstall
     *            The archive list
     * @return The list of shared library names
     */
    private List<String> getComponentNames(final List<File> archivesToUninstall) {
        final List<String> componentStateNames = new ArrayList<String>();

        final List<Component> componentStateHolders = this.systemStateService
                .getComponentStateHolders();
        for (final File archive : archivesToUninstall) {
            for (final Component componentStateHolder : componentStateHolders) {
                if (archive.getName().equals(componentStateHolder.getArchiveName())) {
                    componentStateNames.add(componentStateHolder.getName());
                }
            }
        }

        return componentStateNames;
    }

    /**
     * Get the set of files to install according to the JBI artefact type.
     * 
     * @param jbiType
     *            The expected JBI artefact
     * @return The list of selected JBI artefact files
     * @throws IOException
     * @throws JBIDescriptorException
     * @throws MalformedURLException
     */
    private List<File> getJBIArchives(EntityType jbiType, List<File> archivesToUninstall) {
        final List<File> archives = new ArrayList<File>();

        for (final File archive : archivesToUninstall) {
            try {
                final Jbi descriptor = PackageHelper.loadDescriptor(archive.toURI().toURL());
                if (EntityType.SL_TYPE.equals(jbiType) && descriptor.getSharedLibrary() != null) {
                    archives.add(archive);
                } else if (EntityType.COMPONENT_TYPE.equals(jbiType)
                        && descriptor.getComponent() != null) {
                    archives.add(archive);
                } else if (EntityType.SA_TYPE.equals(jbiType)
                        && descriptor.getServiceAssembly() != null) {
                    archives.add(archive);
                }
            } catch (final PetalsException e) {
                this.log.warning("Unable to read a JBI descriptor", e);
            } catch (final IOException e) {
                this.log.warning("Unable to read a JBI descriptor", e);
            } catch (JBIDescriptorException e) {
                this.log.warning("Unable to read a JBI descriptor", e);
            }
        }

        return archives;
    }

    /**
     * Get the service assembly names from the archive list.
     * 
     * @param archivesToUninstall
     *            The archive list
     * @return The list of service assembly names
     */
    private List<String> getServiceAssemblyNames(final List<File> archivesToUninstall) {
        final List<String> serviceAssemblyNames = new ArrayList<String>();

        final List<ServiceAssembly> serviceAssemblyStateHolders = this.systemStateService
                .getServiceAssemblyStateHolders();
        for (final File archive : archivesToUninstall) {
            for (final ServiceAssembly serviceAssemblyStateHolder : serviceAssemblyStateHolders) {
                if (archive.getName().equals(serviceAssemblyStateHolder.getArchiveName())) {
                    serviceAssemblyNames.add(serviceAssemblyStateHolder.getName());
                    break;
                }
            }
        }

        return serviceAssemblyNames;
    }

    /**
     * Get the shared library names from the archive list.
     * 
     * @param archivesToUninstall
     *            The archive list
     * @return The list of shared library names
     */
    private List<String> getSharedLibraryNames(final List<File> archivesToUninstall) {
        final List<String> sharedLibraryNames = new ArrayList<String>();

        final List<SharedLibrary> sharedLibraryStateHolders = this.systemStateService
                .getSharedLibraryStateHolders();
        for (final File archive : archivesToUninstall) {
            for (final SharedLibrary sharedLibraryStateHolder : sharedLibraryStateHolders) {
                if (archive.getName().equals(sharedLibraryStateHolder.getArchiveName())) {
                    sharedLibraryNames.add(sharedLibraryStateHolder.getName());
                }
            }
        }

        return sharedLibraryNames;
    }

    /**
     * Deploy the specified Archive in the JBI container and start it . Any
     * exception will cause the abortion of the process. A file
     * "fileName.success" will be created if success A file "fileName.failed"
     * will be created otherwise, containing the related exception
     * 
     * @param saFile
     *            the archive file
     * @throws Exception
     */
    private void performDeploySA(File saFile) throws Exception {
        this.log.call();

        final String xml = this.deploymentService.deploy(saFile.toURI().toString());

        if (xml.indexOf(XMLResult.TaskResult.FAILED.toString()) > -1) {
            throw new RuntimeException("Failed to auto-deploy a service assembly");
        } else {
            final String saName = xml.substring(xml.indexOf("<loc-param>") + 11, xml
                    .indexOf("</loc-param>"));
            this.deploymentService.start(saName);
        }
    }

    /**
     * Install the specified Archive in the JBI container and start it . Any
     * exception will cause the abortion of the process. A file
     * "fileName.success" will be created if success A file "fileName.failed"
     * will be created otherwise, containing the related exception
     * 
     * @param componentFile
     *            The component
     * @throws MalformedURLException
     * @throws ReflectionException
     * @throws MBeanException
     * @throws InstanceNotFoundException
     */
    private void performInstallComponent(File componentFile) throws MalformedURLException,
            InstanceNotFoundException, MBeanException, ReflectionException {
        this.log.call();
        final ObjectName newInstallerName = this.installationService.loadNewInstaller(componentFile
                .toURI().toURL().toExternalForm());
        final Object[] nullObjects = new Object[0];
        final String[] nullStrings = new String[0];
        final MBeanServer jmxServer = this.jmxService.getLocalJMXServer();
        final Object result = jmxServer.invoke(newInstallerName, "install", nullObjects,
                nullStrings);
        if (result instanceof ObjectName) {
            final ObjectName componentName = (ObjectName) result;
            jmxServer.invoke(componentName, "start", nullObjects, nullStrings);
        } else {
            throw new RuntimeException(
                    "The install() invocation on the component installer has returned an unexpected object : "
                            + result);
        }
    }

    /**
     * Install the specified shared library in the JBI container Any exception
     * will cause the abortion of the process. A file "fileName.success" will be
     * created if success A file "fileName.failed" will be created otherwise,
     * containing the related exception
     * 
     * @param slFile
     *            The shared library
     */
    private void performInstallSL(File slFile) {
        this.log.call();

        try {
            this.installationService.installSharedLibrary(slFile.toURI().toString());
        } catch (final Throwable e) {
            this.log.error("Failed to auto-install a shared library");
        }
    }

    /**
     * Undeploy the specified service assembly. isToBeDeleted is set to true.
     * 
     * @param name
     *            The service assembly
     */
    private void performUndeploySA(String name) {
        this.deploymentService.forceUndeploy(name);
    }

    /**
     * Uninstall the specified component.
     * 
     * @param name
     *            The component
     */
    private void performUninstallComponent(String name) {
        this.installationService.forceUnloadInstaller(name);
    }

    /**
     * Uninstall a shared library.
     * 
     * @param name
     *            The shared library
     */
    private void performUninstallSL(String name) {
        this.installationService.uninstallSharedLibrary(name);
    }
}
