/**
 * PETALS - PETALS Services Platform. Copyright (c) 2006 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id: DeploySUToComponentask.java 154 6 oct. 06 ofabre $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.jbi.management.task.deployment.deploy;

import java.net.URL;

import javax.jbi.component.ServiceUnitManager;
import javax.jbi.management.DeploymentException;

import org.ow2.petals.container.lifecycle.Installer;
import org.ow2.petals.jbi.descriptor.original.generated.Jbi;
import org.ow2.petals.jbi.descriptor.original.generated.ServiceUnit;
import org.ow2.petals.jbi.management.Context;
import org.ow2.petals.jbi.management.ManagementException;
import org.ow2.petals.jbi.management.admin.AdminService;
import org.ow2.petals.jbi.management.task.AbstractLoggableTask;
import org.ow2.petals.jbi.management.task.deployment.DeploymentUtils;
import org.ow2.petals.util.LoggingUtil;

/**
 * This task deploys service unit to the targeted component <br/>
 * Inputs : SU {@link Jbi}, SU install root URI <br/>
 * Outputs : XML result for of the deployment as {@link String} <br/>
 * The task has got a non-robust approach : it stops at the first subtask failed<br />
 * and have the undo capability.<br />
 * 
 * @author Frederic Gardes - eBM Websourcing
 * @author ofabre - EBM Websourcing
 * 
 */
public class DeploySUToComponentTask extends AbstractLoggableTask {

    /**
     * JBI Container :: JMX Agent Admin
     */
    protected AdminService adminService;

    public DeploySUToComponentTask(LoggingUtil log, AdminService adminService) {
        super(log);
        this.adminService = adminService;
    }

    public void execute(Context context) throws Exception {
        URL installUrl = context.getSuInstallUrl();
        ServiceUnit serviceUnit = context.getSuDescriptorInSa();
        String xmlResult = this.deploySUtoComponent(installUrl, serviceUnit);
        context.setComponentResult(xmlResult);
    }

    /**
     * Deploy the given service unit to the targeted component. The component
     * must be in the STARTED state to allow deployment of service units.
     * 
     * @param installationRoot
     *            the SU installation {@link URL}
     * @param serviceUnit
     *            the service unit production element
     * @return a {@link String} xml result of the service unit deployment on the
     *         targeted component (see section 6.9 of the JSR 208)
     * @throws ManagementException
     *             if component and related object are missing, if the call of
     *             the service unit manager method "deploy" fails or if the
     *             component isn't in the STARTED state
     */
    protected String deploySUtoComponent(URL installUrl, ServiceUnit serviceUnit) throws Exception {
        String xmlReturn = null;
        /*
         * Retrieve the target component life cycle
         */
        String targetedComponent = DeploymentUtils.getServiceUnitTargetedComponent(serviceUnit);
        Installer installer = this.adminService.getInstallerByName(targetedComponent);
        if (installer == null) {
            throw new ManagementException(
                    "The Service Unit can not be deployed, as the component '" + targetedComponent
                            + "' is not installed");
        }

        /*
         * Retrieve the service unit manager of the component that allow to
         * process deployment
         */
        ServiceUnitManager sum = installer.getComponent().getServiceUnitManager();
        if (sum == null) {
            throw new ManagementException("Component '" + targetedComponent
                    + "' doesn't support deployment : service unit manager is null");
        }
        // Set the context class loader to the component's one just for the
        // method invocation
        try {
            Thread.currentThread().setContextClassLoader(sum.getClass().getClassLoader());
            String serviceUnitName = DeploymentUtils.getServiceUnitName(serviceUnit);
            try {
            	xmlReturn = sum.deploy(serviceUnitName, installUrl.getPath());
            } catch(Exception e) {
                try {
		    sum.undeploy(serviceUnitName, installUrl.getPath());
                } catch(Exception ignore) { }
            }	
        } finally {
            Thread.currentThread().setContextClassLoader(this.getClass().getClassLoader());
        }
        return xmlReturn;
    }

    public void undo(Context context) throws Exception {
        this.log.debug("Revert DeploySUToComponent");
        if ((context.getSuInstallUrl() != null) && (context.getSuDescriptorInSa() != null)) {
            try {
                this.undeploySUFromComponent(context.getSuInstallUrl(), context
                        .getSuDescriptorInSa());
            } catch (ManagementException e) {
                this.log.error("Failed to revert a DeploySUToComponentask", e);
            }
        }
    }

    /**
     * Undeploy the given service unit from the targeted component.
     * 
     * @param installUrl
     *            the su installation {@link URL}
     * @param serviceUnit
     *            the service unit production element
     * @throws ManagementException
     *             if component and related object are missing, if the call of
     *             the service unit manager method "undeploy" fails
     */
    private void undeploySUFromComponent(URL installUrl, ServiceUnit serviceUnit)
            throws ManagementException {
        // Retrieve the target component life cycle
        String targetedComponent = DeploymentUtils.getServiceUnitTargetedComponent(serviceUnit);
        if (targetedComponent != null) {
            Installer installer = this.adminService.getInstallerByName(targetedComponent);
            if (installer == null) {
                throw new ManagementException(
                        "The component lifecycle element for service unit deployment must be non null.");
            }
            // Retrieve the service unit manager of the component that allow to
            // process deployment
            ServiceUnitManager sum = installer.getComponent().getServiceUnitManager();
            if (sum == null) {
                throw new ManagementException(
                        "The service unit manager is null for the component '" + targetedComponent
                                + "'. The service units can not be deployed.");
            }
            String serviceUnitName = DeploymentUtils.getServiceUnitName(serviceUnit);
            try {
                Thread.currentThread().setContextClassLoader(sum.getClass().getClassLoader());
                sum.undeploy(serviceUnitName, installUrl.getPath());
            } catch (DeploymentException e) {
                throw new ManagementException("Undeployement of the service unit '"
                        + serviceUnitName + "' has failed (caused by a target component '"
                        + serviceUnitName + "' SU manager exception)", e);
            } finally {
                Thread.currentThread().setContextClassLoader(this.getClass().getClassLoader());
            }
        }
    }
}
